"""
This file contains utilities for control design.
"""
import numpy as np
from scipy.signal import find_peaks

def ctrb(A, B):
    """
    Function to compute the controllability matrix of a system xDot = Ax + Bu
    Inputs:
        A (nxn NumPy Array): A matrix of a system
        B (nxm NumPy Array): B matrix of a system
    Returns:
        [B AB A^2B ... A^(n-1)B]
    """
    #initialize controllabiity matrix as B
    P = B
    for i in range(1, A.shape[0]):
        P = np.hstack((P, np.linalg.matrix_power(A, i) @ B))
    #return controllability matrix
    return P

def obsv(A, C):
    """
    Function to compute the observability matrix of a system xDot = Ax + Bu, y = Cx
    Inputs:
        A (nxn NumPy Array): A matrix of a system
        C (mxn NumPy Array): C matrix of a system
    Returns:
        [C; CA; CA^2; ...; CA^n-1]
    """
    #initialize observability matrix as C
    O = C
    for i in range(1, A.shape[0]):
        P = np.vstack((P, C @ np.linalg.matrix_power(A, i)))
    #return observability matrix
    return O

def is_ctrb(A, B):
    """
    Verify if (A, B) is a controllable pair. Returns true if controllable.
    """
    return np.linalg.matrix_rank(ctrb(A, B)) == A.shape[0]

def is_obsv(A, C):
    """
    Verify if (A, C) is an observable pair. Returns true if observable.
    """
    return np.linalg.matrix_rank(ctrb(A, C)) == A.shape[0]

def calc_cl_poles(A, B, K):
    """
    Function to calculate the closed loop poles of a system xDot = Ax + Bu
    using state feedback with gain matrix K.
    Inputs:
        A (nxn NumPy Array): A matrix
        B (nxm NumPy Array): B Matrix
        K (mxn NumPy Array): Gain matrix
    Returns:
        [lambda1, lambda2, ..., lambdan] (list of floats): closed loop poles of the system with gain K
    """
    return np.linalg.eigvals(A - B @ K).tolist()

def place_poles(A, B, pole_list):
    """
    Function to compute the gain K to place the poles of A - BK at desired posiions using Ackermann's formula.
    This function depends on invertibility of the controllability matrix.
    Inputs:
        A (nxn NumPy Array): A matrix
        B (nxm NumPy Array): B Matrix
        pole_list (list of n complex/float numbers): list of desired pole positions for the closed loop system
    Returns:
        K (mxn NumPy Array): State feedback gain matrix to place the poles of the system in the desired locations
    """
    #find the char. polyn of A
    char_poly = np.poly(np.linalg.eigvals(A))

    #find the desired char. polyn
    char_poly_des = np.poly(pole_list)

    #subtract the desired poles
    rowVec = (char_poly_des[1:] - char_poly[1:]).reshape((1, A.shape[0]))

    #compute the W terms
    Wr = ctrb(A, B)

    #assemble Wr tilda (inverse of toeplitz matrix)
    Wrt = np.eye(A.shape[0])
    for i in range(A.shape[0]):
        for j in range(A.shape[1]):
            if j - i >= 0:
                #check the relation between i and j
                Wrt[i, j] = char_poly[j - i]
            else:
                Wrt[i, j] = 0
    Wrt = np.linalg.pinv(Wrt)

    #return the gain
    return rowVec @ np.linalg.pinv(Wr) @ Wrt

def calc_first_peak(yData):
    """
    Function to find first peak and their indices of a signal yData.
    Inputs:
        yData (m x N NumPy Array): time series output of the system (rows are outputs, columns are at different time steps)
    Returns
        yPeaks (m x 1 NumPy Array): vector of first peak values
        yPeakIndices (m x 1 NumPy Array): vector of indices of the first peaks
    """
    #reshape yData to have two dimensions
    if len(yData.shape) == 1:
        #reshape yData to be a row vector
        yData = yData.reshape((1, yData.shape[0]))

    #initialize an empty peak and peak index array
    yPeaks = []
    yPeakIndices = []
    
    for i in range(yData.shape[0]):
        #slice out the row of outputs
        yiData = yData[i, :]

        #find the indices of the local maxima in yData -> take the first local maximum
        peakIndices, _ = find_peaks(yiData)

        #append the value of y at the first peak index
        yPeaks.append(yiData[peakIndices[0]])

        #append the first peak index
        yPeakIndices.append(peakIndices[0])

    #convert to NumPy Arrays and return
    return np.array([yPeaks]).T, np.array([yPeakIndices]).T

def calc_peak_time(yData, tData):
    """
    Function to calculate the peak time of a system's output. Note: this function 
    will generally only converge for step responses for linear systems. It is designed
    around a typical linear second order response for a stable system.
    Inputs:
        yData (m x N NumPy Array): time series output of the system (rows are outputs, columns are at different time steps)
        tData (1 x N NumPy Array): time series corresponding to yData
    Returns:
        Tp (m x 1 NumPy array): Peak time for each output
    """
    #calculate the indices of the peaks
    _, peakIndices = calc_first_peak(yData)

    #initialize an empty rise time array
    peakTimeArr = []

    #find the times associated with the first peaks
    for i in range(peakIndices.shape[0]):
        #get the time associated with the peak index
        peakTimeArr.append(tData[peakIndices[i, 0]])

    #convert riseTimeArray to a NumPy array and return
    return np.array([peakTimeArr]).T

def calc_ss_value(yData):
    """
    Function to estimate the steady state value of a signal. This will only return a 
    good estimate if the settling time has been reached within tData and the response is stable.
    Returns the vector of final values of the signal.
    Inputs:
        yData (m x N NumPy Array): time series output of the system (rows are outputs, columns are at different time steps)
    Returns
        ySS (m x 1 NumPy Array): vector of final output values
    """
    #reshape yData to have two dimensions
    if len(yData.shape) == 1:
        #reshape yData to be a row vector
        yData = yData.reshape((1, yData.shape[0]))

    #slice out last column in yData and return
    return yData[:, -1].reshape((yData.shape[0], 1))

def find_percent_index(yData, ySS, percent):
    """
    Function to calculate the index at which the output reaches a 
    particular percent of its SS value. Used as a helper function for rise time.
    Moves from the front of the signal to the back. (Data is potentially unordered so cannot binary search)
    Inputs:
        yData (m x N NumPy Array): time series output of the system (rows are outputs, columns are at different time steps)
        ySS (m x 1 NumPy Array): steady state value array
        percent (float): float from 0 to 100 describing desired percentage to find
    Returns:
        indexList (list of m ints): list of indices at which the percentage is achieved
    """
    #reshape yData to have two dimensions
    if len(yData.shape) == 1:
        #reshape yData to be a row vector
        yData = yData.reshape((1, yData.shape[0]))

    #initialize an index list
    indexList = [0]*yData.shape[0]
    for i in range(yData.shape[0]):
        #search from the beginning of the signal to the end for the cutoff
        for j in range(yData.shape[1]):
            #if we pass the perentage threshold, update indexList and break inner loop
            if abs((yData[i, j] - ySS[i, 0])/ySS[i, 0]) > percent/100:
                indexList[i] = j
                break
    #return resulting list of indices
    return indexList

def calc_rise_time(yData, tData):
    """
    Function to calculate the rise time of a system's output, i.e. the time 
    it takes for the response to go from 10% to 90% of its SS value.
    Note: this function will generally only converge for step responses for linear systems. 
    It is designed around a typical linear second order response for a stable system.
    Inputs:
        yData (m x N NumPy Array): time series output of the system (rows are outputs, columns are at different time steps)
        tData (1 x N NumPy Array): time series corresponding to yData
    Returns:
        Tr (m x 1 NumPy array): Rise time for each output
    """    
    #calculate the SS values for each output
    ySS = calc_ss_value(yData)

    #find the 10% and 90% values we're looking for
    indexList10 = find_percent_index(yData, ySS, 10)
    indexList90 = find_percent_index(yData, ySS, 90)

    #initialize an empty rise time array
    TrArr = []

    #find the times associated with each and subtract to get rise time
    for i in range(len(indexList10)):
        TrArr.append(tData[indexList90[i]] - tData[indexList10[i]])

    #return the rise time array
    return np.array([TrArr]).T

def calc_percent_os(yData):
    """
    Function to calculate the percent overshoot of a system's output. Note: this function 
    will generally only converge for step responses for linear systems. It is designed
    around a typical linear second order response for a stable system.
    For this function to work, it is important that y has settled to its SS response at the end of tData.
    Inputs:
        yData (m x N NumPy Array): time series output of the system (rows are outputs, columns are at different time steps)
    Returns:
        %OS (m x 1 NumPy array): Percent overshoot (from 0 to 100) for each output
    """    
    #calculate the peak values
    peakArray, _ = calc_first_peak(yData)

    #calculate the steady state values
    ssArray = calc_ss_value(yData)

    #find the percent overshoot for each element and return
    return 100*np.divide(peakArray - ssArray, ssArray)

def settling_time_helper(y, yss):
    """
    Helper function for settling time. Checks if the +-2% condition is broken for an output value y & SS value yss
    Inputs:
        y (float): scalar output value to evaluate
        yss (float): steady state value
    Returns:
        True/False (boolean): if condition has been broken or not for this particular y
    """
    checkVal = abs((y-yss)/yss)
    return (0.98 <= checkVal) and (checkVal <= 1.02)

def calc_settling_time(yData, tData):
    """
    Function to calculate the settling time of a response, i.e. the time to get and stay within +-2% of the SS value.
    Inputs:
        yData (m x N NumPy Array): time series output of the system (rows are outputs, columns are at different time steps)
        tData (1 x N NumPy Array): time series corresponding to yData
    Returns:
        Ts (m x 1 NumPy array): Settling time for each output
    """
    #reshape yData to have two dimensions
    if len(yData.shape) == 1:
        #reshape yData to be a row vector
        yData = yData.reshape((1, yData.shape[0]))

    #calculate SS values
    ySS = calc_ss_value(yData)

    #initialize an empty settling time array
    TsArr = [0]*ySS.shape[0]

    #search for +-2% index. If none is found, will simply return the final index.
    for i in range(ySS.shape[0]):
        #move backwards through each time series from the end to the beginning
        for j in range(reversed(yData.shape[1])):
            #Add the time to TsArray
            TsArr[i] = tData[j]
            #evaluate if this particular timestep satisfies the +-2% condition
            if not settling_time_helper(yData[i, j], ySS[i]):
                #break the inner loop
                break
    #convert to NumPy array and return
    return np.array([TsArr]).T

def response_info(yData, tData):
    """
    Master function to provide information about the response of the system. Prints and returns
    specifications about the response contained in yData.
    Inputs:
        yData (m x N NumPy Array): time series output of the system (rows are outputs, columns are at different time steps)
        tData (1 x N NumPy Array): time series corresponding to yData
    Returns:
        Tr, Ts, OS, Tp: Rise time, Settling Time, Percent Overshoot, Peak Time NumPy arrays
    """
    #calculate the response specifications
    Tr = calc_rise_time(yData, tData)
    Ts = calc_settling_time(yData, tData)
    Overshoot = calc_percent_os(yData)
    Tp = calc_peak_time(yData, tData)

    #print the response information
    print("Rise Time (s): ", Tr.T)
    print("Settling Time (s):", Ts.T)
    print("Percent Overshoot (%)", Overshoot.T)
    print("Peak Time (s): ", Tp.T)

    #return the response information
    return Tr, Ts, Overshoot, Tp


"""
Utilities file for running simulations. Includes functions to generate initial and goal states.
"""
import numpy as np


#NOTE: DEFINE CONSTANTS FOR INITIAL AND GOAL STATES HERE
#define circle radius
r = 2.5

#define circle center
xC, yC = 2.5, 2.5

#define an initial condition generator
def gen_init_cond(N):
    """
    Function to return initial condition for a system of N bots.
    Evenly spaces a set of turtlebots on a circle.
    """
    #generate a vector of zeros
    q0 = np.zeros((3*N, 1))

    #offset for assymetry to avoid deadlock
    offset = 0.17

    #find the initial condition vector for each turtlebot
    for i in range(N):
        #get x and y, place into the vector
        x0 = r*(np.cos(i/N * 2*np.pi + offset)) + xC
        y0 = r*(np.sin(i/N * 2*np.pi + offset)) + yC
        q0[3*i : 3*i + 3, 0] = np.array([x0, y0, 0])

    return q0

#define a goal state vector generator
def gen_goal_state(N):
    """
    Function to define a goal state vector for a system of N bots.
    Evenly spaces a set of goal states on a circle 180 degrees from the initial cond.
    """
    #generate a vector of zeros
    qD = np.zeros((3*N, 1))

    #find the initial condition vector for each turtlebot
    for i in range(N):
        #get x and y, place into the vector
        x0 = r*(np.cos(i/N * 2*np.pi + np.pi)) + xC
        y0 = r*(np.sin(i/N * 2*np.pi + np.pi)) + yC
        qD[3*i : 3*i + 3, 0] = np.array([x0, y0, 0])
    
    return qD


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
import seaborn as sns

class Dynamics:
    """
    Skeleton class for system dynamics
    Includes methods for returning state derivatives, plots, and animations
    """
    def __init__(self, x0, singleStateDimn, singleInputDimn, f, N = 1):
        """
        Initialize a dynamics object
        Args:
            x0 (N*stateDimn x 1 numpy array): (x01, x02, ..., x0N) Initial condition state vector for all N agents
            stateDimn (int): dimension of state vector for a single agent
            inputDimn (int): dimension of input vector for a single agent
            f (python function): dynamics function in xDot = f(x, u, t) -> This is for a SINGLE instance
            N (int): Number of "agents" in the system, i.e. how many instances we want running in parallel
        """    
        #store the dynamics function
        self._f = f

        #store the number of agents
        self.N = N

        #store the state and input dimensions for the extended system
        self.singleStateDimn = singleStateDimn
        self.singleInputDimn = singleInputDimn
        self.sysStateDimn = singleStateDimn * N
        self.sysInputDimn = singleInputDimn * N

        #store the state and input
        self._x = x0
        self._u = np.zeros((self.sysInputDimn, 1))

        #set the plotting style with seaborn
        sns.set_theme()
        sns.set_context("paper")

    def get_input(self):
        """
        Retrieve the input to the system
        """
        return self._u
    
    def get_state(self):
        """
        Retrieve the state vector
        """
        return self._x
    
    def print_params(self):
        """
        Prints the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        """
        print("Skeleton dynamics class")

    def return_params(self):
        """
        Returns the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        """
        return None
        
    def deriv(self, x, u, t):
        """
        Returns the derivative of the state vector for the extended system
        Args:
            x (sysStateDimn x 1 numpy array): current state vector at time t
            u (sysInputDimn x 1 numpy array): current input vector at time t
            t (float): current time with respect to simulation start
        Returns:
            xDot: state_dimn x 1 derivative of the state vector
        """
        #assemble the derivative vector for the entire system
        xDot = np.zeros((self.sysStateDimn, 1))
        for i in range(self.N):
            #define slicing indices
            stateSlice0 = self.singleStateDimn * i
            stateSlice1 = self.singleStateDimn * (i + 1)
            inputSlice0 = self.singleInputDimn * i
            inputSlice1 = self.singleInputDimn * (i + 1)

            #extract the state and input of the ith agent
            xi = x[stateSlice0  : stateSlice1, 0].reshape((self.singleStateDimn, 1))
            ui = u[inputSlice0 : inputSlice1, 0].reshape((self.singleInputDimn, 1))
            
            #compute the derivative of the ith agent's state vector
            xDot[stateSlice0 : stateSlice1, 0] = self._f(xi, ui, t).reshape((self.singleStateDimn, ))

        #return the assembled derivative vector
        return xDot
    
    def euler_integrate(self, u, t, dt):
        """
        Integrates system dynamics using Euler integration
        Args:
            u (sysInputDimn x 1 numpy array): current input vector at time t
            t (float): current time with respect to simulation start
            dt (float): time step for integration
        Returns:
            x (sysStateDimn x 1 numpy array): state vector after integrating
        """
        #integrate starting at x
        self._x = self.get_state() + self.deriv(self.get_state(), u, t)*dt

        #update the input parameter
        self._u = u

        #return integrated state vector
        return self._x
    
    def rk4_integrate(self, u, t, dt):
        """
        Integrates system dynamics using RK4 integration
        Args:
            u (sysInputDimn x 1 numpy array): current input vector at time t
            t (float): current time with respect to simulation start
            dt (float): time step for integration
        Returns:
            x (sysStateDimn x 1 numpy array): state vector after integrating
        """
        #get current deterministic state
        x = self.get_state()

        #evaluate RK4 constants
        k1 = self.deriv(x, u, t)
        k2 = self.deriv(x + dt*k1/2, u, t + dt/2)
        k3 = self.deriv(x + dt*k2/2, u, t + dt/2)
        k4 = self.deriv(x + dt*k3, u, t + dt)

        #update the state parameter
        self._x = x + dt/6 * (k1 + 2*k2 + 2*k3 + k4)

        #update the input parameter
        self._u = u

        #return the integrated state vector
        return self._x
    
    def integrate(self, u, t, dt, integrator = "rk4"):
        """
        Integrate dynamics with either rk4 or euler integration
        Choose either "rk4" or "euler" to select integrator.
        """
        if integrator == "rk4":
            return self.rk4_integrate(u, t, dt)
        else:
            return self.euler_integrate(u, t, dt)
    
    def show_plots(self, xData, uData, tData, stateLabels = None, inputLabels = None):
        """
        Function to show plots specific to this dynamic system.
        Args:
            xData ((sysStateDimn x N) numpy array): history of N states to plot
            uData ((sysInputDimn x N) numpy array): history of N inputs to plot
            tData ((1 x N) numpy array): history of N times associated with x and u
            stateLabels ((singleStateDimn) length list of strings): Optional custom labels for the state plots
            inputLabels ((singleInputDimn) length list of strings): Optional custom labels for the input plots
        """
        #Plot each state variable in time
        fig, axs = plt.subplots(self.singleStateDimn + self.singleInputDimn)
        fig.suptitle('Evolution of States and Inputs in Time')
        xlabel = 'Time (s)'

        #set state and input labels
        if stateLabels is None:
            stateLabels = ["X" + str(i + 1) for i in range(self.singleStateDimn)]
        if inputLabels is None:
            inputLabels = ["U" + str(i + 1) for i in range(self.singleInputDimn)]

        #plot the states for each agent
        for j in range(self.N):
            n = 0 #index in the subplot
            for i in range(self.singleStateDimn):
                axs[n].plot(tData.reshape((tData.shape[1], )).tolist()[0:-1], xData[self.singleStateDimn*j + i, :].tolist()[0:-1])
                axs[n].set(ylabel=stateLabels[n]) #pull labels from the list above
                axs[n].grid()
                n += 1

            #plot the inputs
            for i in range(self.singleInputDimn):
                axs[i+self.singleStateDimn].plot(tData.reshape((tData.shape[1], )).tolist()[0:-1], uData[self.singleInputDimn*j + i, :].tolist()[0:-1])
                axs[i+self.singleStateDimn].set(ylabel=inputLabels[i])
                axs[i+self.singleStateDimn].grid()
        
        axs[self.singleStateDimn + self.singleInputDimn - 1].set(xlabel = xlabel)
        legendList = ["Agent " + str(i) for i in range(self.N)]
        plt.legend(legendList)
        plt.show()
    
    def show_animation(self, xData, uData, tData, axis_lims, labels, anim_point, anim_line = None, animate = True, obsManager = None):
        """
        Function to play animations specific to this dynamic system.
        Args:
            xData ((sysStateDimn x N) numpy array): history of N states to plot
            uData ((sysInputDimn x N) numpy array): history of N inputs to plot
            tData ((1 x N) numpy array): history of N times associated with x and u
            axis_lims (Python List, length 4): list of axis limits, ex: [-0.25, 5.25, -0.25, 5.25]
            labels (Python list, length 3): List of strings ["Xlabel", "Ylabel", "Title"] for plot
            anim_point (Python function): function to be called at an index i, returns x, y to be animated
            anim_line (Python function): function to be called at an index i, returns line to be animated
            obsManager (ObstacleManager object): If included, will animate the obstacles in the scene.
        """
        #Set constant animtion parameters
        FREQ = 50 #control frequency, same as data update frequency
        
        if animate:
            fig, ax = plt.subplots()
            # set the axes limits
            ax.axis(axis_lims)
            # set equal aspect such that the circle is not shown as ellipse
            ax.set_aspect("equal")

            # create a set of points in the axes
            point, = ax.plot([],[], marker="o", linestyle='None')

            #plot the obstacles if present
            if obsManager is not None:
                #plot the circular obstacles
                for i in range(obsManager.NumObs):
                    #get the obstacle 
                    obsI = obsManager.get_obstacle_i(i)
                    center = obsI.get_center()

                    #plot obstacle on YZ
                    plt.gca().add_patch(plt.Circle((center[1, 0], center[2, 0]), radius = obsI.get_radius(), fc = 'c'))

            #check for anim_line
            if anim_line is not None:
                #define the line for the quadrotor
                line, = ax.plot([], [], 'o-', lw=2)

            def anim_func(i):
                #Call animate_func to get the x and y to plot
                x, y = anim_point(i)

                #set the data to points and return points
                point.set_data(x, y)

                #check for anim_line
                if anim_line is not None:
                    #call the line animation function
                    thisx, thisy = anim_line(i)
                    #set the data for the line
                    line.set_data(thisx, thisy)

                    #if we use a line, return line and point
                    return line, point
                else:
                    #if no line is present, just return the point
                    return point,

            num_frames = xData.shape[1]-1

            anim = animation.FuncAnimation(fig, anim_func, frames=num_frames, interval=1/FREQ*1000, blit=True)

            plt.xlabel(labels[0])
            plt.ylabel(labels[1])
            plt.title(labels[2])
            plt.show()
    
    
"""
**********************************
PLACE YOUR DYNAMICS FUNCTIONS HERE
**********************************
"""

class DoubleIntegratorDyn(Dynamics):
    """
    Double Integrator System
    May initialize N integrators to run in parallel.
    """
    def __init__(self, x0, N = 1):
        #define the double integrator dynamics
        def double_integrator(x, u, t):
            """
            Double integrator dynamics
                x (2x1 NumPy array): state vector
                u (1x1 NumPy array): input vector
                t (float): current time
            """
            return np.array([[0, 1], [0, 0]]) @ x + np.array([[0, 1]]).T @ u
        super().__init__(x0, 2, 1, double_integrator, N = N)
    
    def print_params(self):
        """
        Prints the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        """
        print("Double Integrator Dynamics")

class MSDRamp(Dynamics):
    """
    Mass-spring-damper system on a ramp.
    """
    def __init__(self, x0, m = 0.5, g = 9.81, k = 15, b = 0.5, theta = np.pi/6, N = 1):
        """
        Inputs:
            x0 (2x1 NumPy Array): Initial condition
            m (float): mass in kg
            g (float): acceleration due to gravity (m/s^2)
            k (float): spring constant (N/m)
            b (float): damping constant (N/(m/s))
            theta (float): angle of ramp
            N (int): number of agents
        """
        self.m = m
        self.g = g
        self.k = k
        self.b = b
        self.theta = theta

        def msd_ramp(x, u, t):
            """
            Mass spring damper ramp dynamics
            Inputs:
                x (2x1 NumPy array): current state vector
                u (1x1 NumPy Array): force applied to mass (typically 0)
                t (float): current time in simulation
            """
            return np.array([[x[1, 0]], [u[0, 0]/m -k/m * x[0, 0] - b/m * x[1, 0] - g*np.sin(theta)]])
        
        #call the init function on the MSD ramp system
        super().__init__(x0, 2, 1, msd_ramp, N = N)

    def print_params(self):
        """
        Prints the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        """
        print("Mass-Spring-Damper System")
        print("Mass: ", self.m)
        print("Gravitational accel: ", self.g)
        print("Spring constant: ", self.k)
        print("Damping constant: ", self.b)
        print("Ramp angle: ", self.theta)

    def return_params(self):
        """
        Returns the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        Returns:
            self.m, self.g, self.k, self.b, self.theta
        """
        return self.m, self.g, self.k, self.b, self.theta

class DoublePendulum(Dynamics):
    """
    Double pendulum dynamics
    """
    def __init__(self, x0, m = 0.5, I = 0.5, L = 1, g = 9.81, N = 1):
        """
        Inputs:
            x0: initial condition [q1, q2, q1Dot, q2Dot]
            m: mass of joints
            I: inertia of joints
            L: length of joints
            g: gravitational acceleration
        """
        #store the system parameters
        self.m = m
        self.I = I
        self.L = L
        self.g = g

        def double_pend(x, u, t):
            """
            Double pendulum dynamics
            Includes a torque input at each joint.
            """
            #extract the states
            q1, q2, q1Dot, q2Dot = x[0, 0], x[1, 0], x[2, 0], x[3, 0]

            #compute the inertia matrix
            M = np.array([[2*I + 1.5*m*L**2 + m*L**2*np.cos(q2), I+0.25*m*L**2+0.5*m*L**2*np.cos(q2)],
                          [I+0.25*m*L**2+0.5*m*L**2*np.cos(q2), I+0.25*m*L**2]])
            #compute C
            C = np.array([[0, -m*L**2*q1Dot*np.sin(q2)-0.5*m*L**2*q2Dot*np.sin(q2)], [0.5*m*L**2*q1Dot*np.sin(q2), 0]])

            #compute N
            N = np.array([[0.5*m*g*L*(np.sin(q1+q2)+3*np.sin(q1)), 0.5*m*g*L*np.sin(q1+q2)]]).T

            #compute qDDot
            qDot = np.array([[q1Dot, q2Dot]]).T
            qDDot = np.linalg.inv(M)@(u - C@qDot - N)

            #return [qDot, qDDot]
            return np.vstack((qDot, qDDot))
        
        #call the init function on the double pendulum system
        super().__init__(x0, 4, 2, double_pend, N = N)
    
    def print_params(self):
        """
        Prints the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        """
        print("Double Pendulum System")
        print("Mass: ", self.m)
        print("Inertia: ", self.I)
        print("Length: ", self.L)
        print("Gravitational accel: ", self.g)

    def return_params(self):
        """
        Returns the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        Returns:
            self.m, self.I, self.L, self.g
        """
        return self.m, self.I, self.L, self.g


class FlywheelPendulum(Dynamics):
    """
    Pendulum driven by a flywheel
    """

class MarsProbe(Dynamics):
    """
    Dynamics of a particle falling through the atmosphere of mars.
    """
    def __init__(self, x0, m = 572.743, Cd = 1.46, A = 5.5155, ftMax = 9806.65, N = 1):
        """
        Init function for mars probe.
        
        Mission Phases and Probe Parameters:
        - https://ntrs.nasa.gov/api/citations/20080034645/downloads/20080034645.pdf
        - Enter atmosphere at 5600 m/s
        - Chute deployment occurs at Mach 1.65, 12.9 km above surface
        - Touchdown occurs at 0.7 m/s
        
        Inputs:
            x0 (2x1 NumPy Array): position vector of the system
            m (float): mass of the probe
            Cd (float): drag coefficient of probe 
                NOTE: Cd for a parachute ~0.5 -> add this to Cd for just probe. Chute area ~200m^2.
                      https://downloads.spj.sciencemag.org/space/2022/9805457.pdf
            A (float): surface area for drag calculation
            ftMax (float): maximum thrust force of probe
            N (int): number of probes to simulate
        """
        #properties of probe
        self.m = m
        self.Cd = Cd
        self.ftMax = ftMax

        #properties of Mars
        self.rho = 0.02 #density of atmosphere (kg/m^3)
        self.M = 6.41693 * 10**23 #mass of Mars (kg)
        self.R = 3390 * 10**3 #radius of Mars (m)
        self.G = 6.6743 * 10**(-11) #universal gravitational constant

        def probe_dyn(x, u, t):
            """
            Mars probe dynamics.
            Inputs:
                x (2x1 NumPy Array): [y, yDot] for y distance to the center of mars
                u (1x1 NumPy Array): thrust force of the probe
                t (float): time
            """
            y = x[0, 0]
            yDot = x[1, 0]
            yDDot = -self.G * self.M / (y**2) + 0.5*self.rho*self.Cd*yDot**2/self.m + u[0, 0]/m
            return np.array([[yDot, yDDot]]).T
        
        #call the init function on the probe pendulum system
        super().__init__(x0, 2, 1, probe_dyn, N = N)

class TurtlebotSysDyn(Dynamics):
    """
    System of N Turtlebots
    """
    def __init__(self, x0, N = 1, rTurtlebot = 0.15):
        """
        Init function for a system of N turtlebots.
        Args:
            x0 (NumPy Array): (x1, y1, phi1, ..., xN, yN, phiN) initial condition for all N turtlebots
            N (Int, optional): number of turtlebots in the system
            rTurtlebot (float): radius of the turtlebots in the system
        """

        #define the turtlebot dynamics
        def f_turtlebot(x, u, t):
            #extract the orientation angle of the Nth turtlebot
            PHI = x[2, 0]
            return np.array([[np.cos(PHI), 0], [np.sin(PHI), 0], [0, 1]])@u

        #call the super init function to create a turtlebot system
        super().__init__(x0, 3, 2, f_turtlebot, N)

        #store a copy of the augmented input vector for feedback linearization
        self._z = np.zeros((self.sysInputDimn, 1))

        #store the turtlebot radius
        self.rTurtlebot = rTurtlebot 

    def print_params(self):
        """
        Prints the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        """
        print("Turtlebot System")
        print("Turtlebot radius: ", self.rTurtlebot)

    def return_params(self):
        """
        Returns the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        Returns:
            self.rTurtlebot
        """
        return self.rTurtlebot

    def set_z(self, z, i):
        """
        Function to set the value of z, the augmented input vctor.
        Inputs:
            z ((2N x 1) NumPy Array): Augmented input vector
            i (int): index we wish to place the updated z at
        """
        #store in class attribute
        self._z[2*i : 2*i + 2, 0] = z.reshape((2, ))
    
    def get_z(self):
        """
        Function to return the augmented input vector, z, at any point.
        """
        #retrieve and return augmented input vector
        return self._z
    
    def show_plots(self, xData, uData, tData):
        #Plot the spatial trajectory of the turtlebots
        fig, ax = plt.subplots()
        ax.set_aspect("equal")
        #iterate over each turtlebot state vector
        for j in range(self.N):
            xCoords = xData[3*j, :].tolist() #extract all of the velocity data to plot on the y axis
            yCoords = xData[3*j+1, :].tolist() #remove the last point, get artefacting for some reason
            ax.plot(xCoords[0:-1], yCoords[0:-1])
        
        legendList = ["Agent " + str(i) for i in range(self.N)]
        plt.legend(legendList)
        plt.xlabel("X Position (m)")
        plt.ylabel("Y Position (m)")
        plt.title("Positions of Turtlebots in Space")
        plt.show()

        #call the super plots with custom labels to show the individual states
        stateLabels = ['X Pos (m)', 'Y Pos (m)', 'Phi (rad)']
        inputLabels = ["V (m/s)", "Omega (rad/s)"]
        super().show_plots(xData, uData, tData, stateLabels, inputLabels)
    
    def show_animation(self, xData, uData, tData, animate = True, obsManager = None):
        """
        Shows the animation and visualization of data for this system.
        Args:
            xData (stateDimn x N Numpy array): state vector history array
            u (inputDimn x N numpy array): input vector history array
            t (1 x N numpy array): time history
            animate (bool, optional): Whether to generate animation or not. Defaults to True.
            obsManager (ObstacleManager): if included, will animate any obstacles present
        """
        #define animation function to return the x, y points to be animated
        def anim_point(i):
            x = []
            y = []
            #get the x, y data associated with each turtlebot
            for j in range(self.N):
                x.append(xData[3*j, i])
                y.append(xData[3*j+1, i])
            return x, y
    
        #define axis limits
        axis_lims = [-0.25, 5.25, -0.25, 5.25]

        #define labels
        axis_labels = ["X Position (m)", "Y Position (m)", "Positions of Turtlebots in Space"]

        #call the super() animation function
        super().show_animation(xData, uData, tData, axis_lims, axis_labels, anim_point, anim_line = None, animate = animate, obsManager = obsManager)



class PlanarQrotor(Dynamics):
    def __init__(self, x0 = np.zeros((8, 1)), m = 0.92, Ixx = 0.0023, l = 0.15, N = 1):
        """
        Init function for a Planar quadrotor system.
        State Vector: X = [x, y, z, theta, x_dot, y_dot, z_dot, theta_dot]
        Input Vector: U = [F, M]
        
        Args:
            x0 ((8 x 1) NumPy Array): initial state (x, y, z, theta, x_dot, y_dot, z_dot, theta_dot)
            m (float): mass of quadrotor in kg
            Ixx (float): moment of inertia about x axis of quadrotor
            l (float): length of one arm of quadrotor
            N (int): number of quadrotors
        """
        #store physical parameters
        self._m = m
        self._Ixx = Ixx
        self._g = 9.81
        self._l = l

        #define quadrotor dynamics
        def quadrotor_dyn(X, U, t):
            """
            Returns the derivative of the state vector
            Args:
                X (8 x 1 numpy array): current state vector at time t
                U (2 x 1 numpy array): current input vector at time t
                t (float): current time with respect to simulation start
            Returns:
                xDot: state_dimn x 1 derivative of the state vector
            """
            #unpack the input vector
            F, M = U[0, 0], U[1, 0]
            F = max(0, F) #Cut off force at zero to prevent negative thrust
            
            #unpack the state vector
            x_dot, y_dot, z_dot = X[4, 0], X[5, 0], X[6, 0] #velocities
            theta, theta_dot = X[3, 0], X[7, 0] #orientations
            
            #calculate the second time derivatives of each
            x_ddot = 0
            y_ddot = (-np.sin(theta)*F)/self._m
            z_ddot = (np.cos(theta)*F - self._m*self._g)/self._m
            theta_ddot = M/self._Ixx
            
            #construct and return state vector        
            return np.array([[x_dot, y_dot, z_dot, theta_dot, x_ddot, y_ddot, z_ddot, theta_ddot]]).T
        
        #call the super init function
        super().__init__(x0, 8, 2, quadrotor_dyn, N = N)

    def print_params(self):
        """
        Prints the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        """
        print("Planar Quadrotor System")
        print("Mass: ", self._m)
        print("Inertia: ", self._Ixx)
        print("Gravitational accel: ", self._g)
        print("Arm length: ", self._l)

    def return_params(self):
        """
        Returns the parameters of the dynamics object.
        e.g. mass, gravitational acceleration, ...
        Returns:
            self._m, self._Ixx, self._g, self._l
        """
        return self._m, self._Ixx, self._g, self._l

    def show_plots(self, xData, uData, tData):
        #Plot each state variable in time
        stateLabels = ['X Pos (m)', 'Y Pos (m)', 'Z Pos (m)', 'Theta (rad)', 'X Vel (m/s)', 'Y Vel (m/s)', 'Z Vel (m/s)', 'Angular Vel (rad/s)']
        inputLabels = ['Force (N)', 'Moment (N*m)']
        super().show_plots(xData, uData, tData, stateLabels, inputLabels)
        
    def show_animation(self, xData, uData, tData, animate = True, obsManager = None):
        """
        Shows the animation and visualization of data for this system.
        Args:
            xData (stateDimn x N Numpy array): state vector history array
            u (inputDimn x N numpy array): input vector history array
            t (1 x N numpy array): time history
            animate (bool, optional): Whether to generate animation or not. Defaults to True.
            obsManager (ObstacleManager): Manager object for obstacles. If included, will animate the obstacles.
        """
        #Set constant animtion parameters
        FREQ = 50 #control frequency, same as data update frequency
        L = self._l #quadrotor arm length

        #define point animation function
        def anim_point(i):
            y = xData[1, i]
            z = xData[2, i]
            return y, z
        
        #define line animation function
        def anim_line(i):
            y = xData[1, i]
            z = xData[2, i]

            #draw the quadrotor line body
            theta = xData[3, i]
            x1 = y + L*np.cos(theta)
            x2 = y - L*np.cos(theta)
            y1 = z + L*np.sin(theta)
            y2 = z - L*np.sin(theta)
            thisx = [x1, x2]
            thisy = [y1, y2]
            return thisx, thisy
        
        #define axis limits
        axis_lims = [-1, 2.5, 0, 2.5]

        #define plot labels
        labels = ["Y Position (m)", "Z Position (m)", "Position of Drone in Space"]

        #call the super animation function
        super().show_animation(xData, uData, tData, axis_lims, labels, anim_point, anim_line, animate = animate, obsManager=obsManager)


"""
Class containing custom errors for the simulator
"""

class ShapeError(Exception):
    """
    Exception raised for wrong size passed into hat map
    """
    def __init__(self):
        super().__init__("Shape of input vector is incorrect")

class NumJointsError(Exception):
    """
    Exception raised for incorrect theta array
    """
    def __init__(self):
        super().__init__("Input does not match number of joints")


"""
This file contains utilities and classes associated with 
the kinematics of manipulator arms.
"""
import numpy as np
from .exceptions import *
from .transforms import *

class Manipulator:
    """
    Class for a manipulator arm. Implements product of exponentials, inverse kinematics,
    jacobian computations.
    """
    def __init__(self, twistList, gst0):
        """
        Init function for classes.
        Inputs:
            twistList (List of Twist objects): List of twists in order from spatial to tool frame
            gst0 (4x4 NumPy Array): base configuration transformation
        """

        #store the twist list
        self._twist_list = twistList
        #store the number of joints
        self.n = len(self._twist_list)
        #store the base transformation
        self._gst0 = gst0

    def compute_fk(self, theta):
        """
        Compute the forward kinematics map of the manipulator
        using the product of exponentials.
        Inputs:
            theta (list of floats): List of joint angles in order from spatial to tool frame
        Returns:
            gst(theta) (4x4 NumPy array): SE(3) transformation from spatial to tool frame
        """
        #check length of input
        if len(theta) != self.n:
            raise NumJointsError()

        #compute the product of the n exponentials
        gst_theta = calc_poe(self._twist_list, theta)

        #multiply by base transformation and return
        return gst_theta @ self._gst0
    
    def compute_xi_i_prime(self, i, theta):
        """
        compute xi_i' for use in the spatial jacobian
        Inputs:
            i (int): index (starting from 1) of the twist we with to transform
            theta (list of floats): list of angles
        """
        #compute the product of exponentials up to i-1
        g_im1 = calc_poe(self._twist_list[0:i-1], theta[0:i-1])
        Ad_gim1 = calc_adjoint(g_im1)

        #multiply the adjoint by xi_i and return
        xi_i = self._twist_list[i-1].get_xi()
        return Ad_gim1 @ xi_i
    
    def compute_spatial_jacobian(self, theta):
        """
        Compute the spatial jacobian of a manipulator arm at 
        configuration theta = [theta1, theta2, ..., thetan]
        Inputs:
            theta (list of floats): List of joint angles in order from spatial to tool frame
        Returns:
            Jst_s(theta) (6xn NumPy Array): Spatial manipulator jacobian
        """
        #check length of input
        if len(theta) != self.n:
            raise NumJointsError()

        #initialize Jst_s with xi_1
        Jst_s = self._twist_list[0].get_xi()

        #loop over the remaining twists
        for i in range(1, self.n):
            #compute xi_i' - remember to shift to 1-indexing
            xi_i_prime = self.compute_xi_i_prime(i+1, theta)
            #update the spatial jacobian
            Jst_s = np.hstack((Jst_s, xi_i_prime))

        #return the spatial jacobian
        return Jst_s
    
    def compute_body_jacobian(self, theta):
        """
        Compute the body jacobian of a manipulator arm at 
        configuration theta = [theta1, theta2, ..., thetan]
        Inputs:
            theta (list of floats): List of joint angles in order from spatial to tool frame
        Returns:
            Jst_b(theta) (6xn NumPy Array): Body manipulator jacobian
        """
        #calculate the transformation gst(theta) and its adjoint
        gst_theta = self.compute_fk(theta)
        ad_gst = calc_adjoint(gst_theta)

        #calculate the spatial jacobian and extract the body jacobian from the adjoint
        Jst_b = np.linalg.inv(ad_gst)@self.compute_spatial_jacobian(theta)
        return Jst_b


import numpy as np
from .state_estimation import *

"""
File containing controllers 
"""
class Controller:
    def __init__(self, observer, lyapunovBarrierList = None, trajectory = None, depthCam = None):
        """
        Skeleton class for feedback controllers
        Args:
            observer (Observer): state observer object
            lyapunov (List of LyapunovBarrier): list of LyapunovBarrier objects
            trajectory (Trajectory): trajectory for the controller to track (could just be a constant point!)
            depthCam (PlanarDepthCam or Lidar): depth camera object
        """
        #store input parameters
        self.observer = observer
        self.lyapunovBarrierList = lyapunovBarrierList
        self.trajectory = trajectory
        self.depthCam = depthCam
        
        #store input
        self._u = np.zeros((self.observer.singleInputDimn, 1))
    
    def eval_input(self, t):
        """
        Solve for and return control input
        Inputs:
            t (float): time in simulation
        Returns:
            u ((Dynamics.singleInputDimn x 1)): input vector, as determined by controller
        """
        self._u = np.zeros((self.observer.singleInputDimn, 1))
        return self._u
    
    def get_input(self):
        """
        Retrieves input stored in class parameter
        Returns:
            self._u: most recent input stored in class paramter
        """
        return self._u
    
class FFController(Controller):
    """
    Class for a simple feedforward controller.
    """
    def __init__(self, observer, lyapunovBarrierList = None, trajectory = None, depthCam = None):
        """
        Init function for a ff controller
        Args:
            observer (Observer): state observer object
            ff (NumPy Array): constant feedforward input to send to system
            lyapunov (List of LyapunovBarrier): list of LyapunovBarrier objects
            trajectory (Trajectory): trajectory for the controller to track (could just be a constant point!)
        """
        self.ff = np.array([[9.81*0.92, 0]]).T
        self.depthCam = depthCam
        super().__init__(observer, lyapunovBarrierList, trajectory)
    
    def eval_input(self, t):
        """
        Solve for and return control input
        Inputs:
            t (float): time in simulation
        Returns:
            u ((Dynamics.singleInputDimn x 1)): input vector, as determined by controller
        """
        if self.depthCam is not None:
            #test the depth camera
            self.depthCam.get_pointcloud()

        self._u = self.ff
        return self._u

class ControllerManager(Controller):
    def __init__(self, observerManager, ControlType, barrierManager = None, trajectoryManager = None, depthManager = None):
        """
        Managerial class that points to N controller instances for the system. Interfaces
        directly with the overall system dynamics object.
        Args:
            observerManager (ObserverManager)
            barrierManager (BarrierManager)
            trajectoryManager (TrajectoryManager)
            depthManager (LidarManager or DepthManager)
            ControlType (Class Name): Name of a class for a controller
        """
        #store input parameters
        self.observerManager = observerManager
        self.barrierManager = barrierManager
        self.trajectoryManager = trajectoryManager
        self.depthManager = depthManager
        self.ControlType = ControlType

        #store the input parameter (should not be called directly but with get_input)
        self._u = None

        #get the number of agents in the system
        self.N = self.observerManager.dynamics.N

        #create a controller dictionary
        self.controllerDict = {}

        #create N separate controllers - one for each agent
        for i in range(self.N):
            #extract the ith trajectory
            try:
                trajI = self.trajectoryManager.get_traj_i(i)
            except:
                print("No trajectory passed in.")
                trajI = None

            #get the ith observer object
            try:
                egoObsvI = self.observerManager.get_observer_i(i)
            except:
                print("No state observer passed in.")
                egoObsvI = None

            #get the ith barrier object
            try:
                barrierI = self.barrierManager.get_barrier_list_i(i)
            except:
                print("No barrier/lyapunov object passed in.")
                barrierI = None

            #get the ith depth camera object
            try:
                depthI = self.depthManager.get_depth_cam_i(i)
            except:
                print("No depth camera/LIDAR object passed in.")
                depthI = None

            #create a controller of the specified type
            self.controllerDict[i] = self.ControlType(egoObsvI, barrierI, trajI, depthI)

    def eval_input(self, t):
        """
        Solve for and return control input for all N agents. Solves for the input ui to 
        each agent in the system and assembles all of the input vectors into a large 
        input vector for the entire system.
        Inputs:
            t (float): time in simulation
        Returns:
            u ((Dynamics.sysInputDimn x 1)): input vector, as determined by controller
        """
        #initilialize input vector as zero vector - only want to store once all have been updated
        u = np.zeros((self.observerManager.dynamics.sysInputDimn, 1))
        singleInputDimn = self.observerManager.dynamics.singleInputDimn

        #loop over the system to find the input to each agent
        for i in range(self.N):
            #solve for the latest input to agent i, store the input in the u vector
            self.controllerDict[i].eval_input(t)
            u[singleInputDimn*i : singleInputDimn*(i + 1)] = self.controllerDict[i].get_input()

        #store the u vector in self._u
        self._u = u

        #return the full input vector
        return self._u
    
    def get_input(self):
        """
        Retrieves input stored in class parameter
        Returns:
            self._u: most recent input stored in class paramter
        """
        return self._u


"""
This file contains the core functionalities for the different Depth Cameras/LIDAR Sensors
"""
import numpy as np

class PlanarDepthCam:
    def __init__(self, index, observerManager, obstacleManager, mean = None, sd = None):
        """
        Init function for a quadrotor depth camera. Returns an XYZ pointcloud of the 
        positions of the obstacles in the environment in the frame of the ith quadrotor.
        Note that the pointcloud is generated ONLY in the YZ plane due to the planar nature
        of the system.

        Inputs:
            observerManager (ObserverManager): manager for state estimation of ith quadrotor
            obstacleManager (ObstacleManager): manager for the set of NumObs obstacles
            mean (float): mean for measurement noise
            sd (float): standard deviation for measurement noise
        """
        #store input paramters
        self.index = index
        self.obstacleManager = obstacleManager
        self.mean = mean
        self.sd = sd

        #extract the observer of quadrotor i
        self.observer = observerManager.get_observer_i(self.index)

        #store an attribute for the number of points to be sampled per obstacle
        self.numPts = 1000//self.obstacleManager.NumObs

        #store an attribute for the pointcloud data
        self._ptcloudData = {}

    def calc_ptcloud_world(self):
        """
        Calculate the pointcloud of the environment using the obstalce manager
        This is calculated in the world frame. As this is a planar camera
        the x coordinate is always zero - pointcloud is generated in YZ plane.
        Inputs:
            None
        Returns:
            3xN NumPy array: pointcloud of environment in the world frame
        """
        #initialize arrays for each coordinate
        xList = []
        yList = []
        zList = []
        for i in range(self.obstacleManager.NumObs):
            #extract the ith obstacle object
            obsI = self.obstacleManager.get_obstacle_i(i)

            #get the center (XYZ) and radius
            qCenter, r = obsI.get_center(), obsI.get_radius()

            #calculate a sample of points on the obstacle in the YZ plane
            for j in range(self.numPts):
                #calculate an angle theta to sample around the circle
                thetaJ = j*2*np.pi/self.numPts

                #calculate each coordinate
                xList.append(0)
                yList.append(qCenter[1, 0] + r*np.cos(thetaJ))
                zList.append(qCenter[2, 0] + r*np.sin(thetaJ))

        #assemble and return the pointcloud in the world frame
        return np.array([xList, yList, zList])
    
    def compute_rotation(self, theta):
        """
        Compute the rotation matrix from the quadrotor frame to the world frame
        Inputs:
            theta: angle of rotation about the x-axis
        Returns:
            Rsq: rotation matrix between spatial and quadrotor frames
        """
        Rsq = np.array([[1, 0, 0], 
                        [0, np.cos(theta), -np.sin(theta)], 
                        [0, np.sin(theta), np.cos(theta)]])
        return Rsq
    
    def calc_ptcloud(self):
        """
        Calculates the pointcloud in the frame of quadrotor i, updates the 
        self._ptcloudData attribute.
        """
        #use the observer to get the position and angle
        state = self.observer.get_state() #(x, y, z, theta, x_dot, y_dot, z_dot, theta_dot)

        #get position and angle of the quadrotor
        q, theta = state[0:3].reshape((3, 1)), state[3, 0]

        #compute rotation matrix to spatial frame
        Rsq = self.compute_rotation(theta)

        #compute pointcloud in world frame
        ptcloudWorld = self.calc_ptcloud_world()

        #transform pointcloud
        ptcloudQrotor = Rsq.T @ (ptcloudWorld - q)

        #store in the pointcloud attribute
        self._ptcloudData["ptcloud"] = ptcloudQrotor
        self._ptcloudData["stateVec"] = state
        return self._ptcloudData
    
    def get_pointcloud(self, update = True):
        """
        Returns the pointcloud dictionary from the class attribute 
        Args:
            update: whether or not to recalculate the pointcloud
        Returns:
            Dict: dictionary of pointcloud points and state vector at time of capture
        """
        #first, calculate the pointcloud
        if update:
            self.calc_ptcloud()
        return self._ptcloudData

class EgoLidar:
    def __init__(self, index, observerManager, mean = None, sd = None):
        """
        Lidar for an ego turtlebot. Returns a pointcloud containing
        other turtlebots in the environment.

        Args:
            index (int) : index of the ego turtlebot (zero-indexed from the first turtlebots)
            observerManager (ObserverManager) : Manager object for state estimation
            mean (float): mean for measurement noise
            sd (float): standard deviation for measurement noise
        """
        #store input parameters
        self.index = index
        self.observerManager = observerManager
        self.mean = mean
        self.sd = sd

        #store an attribute for the pointcloud
        self.numPts = 20 #number of points per bot for depth camera observation
        self._ptcloudData = {} #pointcloud attribute (initialize far away)

    def calc_orientation(self):
        """
        Function to calculate the orientation of the turtlebot WRT the world frame.
        Args:
            None
        Returns:
            Rse: rotation matrix from ego frame to world frame
        """
        qo = (self.observerManager.get_observer_i(self.index)).get_state()
        phi = qo[2, 0]
        Rse = np.array([[np.cos(phi), -np.sin(phi), 0], [np.sin(phi), np.cos(phi), 0], [0, 0, 1]])
        return Rse

    def calc_ptcloud(self):
        """
        Function to compute the pointcloud of the environment from the ego turtlebot frame.
        Args:
            None
        Returns:
            ptcloud (3 x N NumPy Array): pointcloud containing (x, y, z) coordinates of obstacles within ego frame.
        """
        #get the ego state vector and orientation
        qEgo = (self.observerManager.get_observer_i(self.index)).get_state()
        pse = np.array([[qEgo[0, 0], qEgo[1, 0], 0]]).T #ego XYZ position
        Rse = self.calc_orientation() #ego rotation matrix to spatial frame

        #get a list of observers for all obstacle turtlebots
        obsIndexList = list(self.observerManager.observerDict.keys()) #get a list of all turtlebot indices
        obsIndexList.remove(self.index) #remove the ego index

        #define an array of angles to generate points over
        thetaArr = np.linspace(0, 2*np.pi, self.numPts).tolist()

        #define an empty pointcloud - should be 3 x numPts
        ptcloud = np.zeros((3, len(obsIndexList) * self.numPts))
        
        #define a number of iterations
        j = 0

        #iterate over all obstacle turtlebots
        for i in obsIndexList:
            #get the state vector and radius of the ith turtlebot
            qo = (self.observerManager.get_observer_i(i)).get_state()
            rt = self.observerManager.dynamics.rTurtlebot

            #calculate the points on the circle - all z values are zero
            xList = [rt*np.cos(theta) + qo[0, 0] for theta in thetaArr]
            yList = [rt*np.sin(theta) + qo[1, 0] for theta in thetaArr]
            zList = [0 for theta in thetaArr]

            #put the points in a numpy array
            ptcloudI = np.array([xList, yList, zList])

            #transform the points into the ego frame
            ptcloudIego = (np.linalg.inv(Rse)@(ptcloudI - pse))

            #store in the ptcloud
            ptcloud[:, j * self.numPts : j*self.numPts + self.numPts] = ptcloudIego

            #increment number of iterations
            j += 1
            
        #store both the pointcloud and the ego state vector at the time the pointcloud is taken
        self._ptcloudData["ptcloud"] = ptcloud
        self._ptcloudData["stateVec"] = qEgo
        return self._ptcloudData
        
    def get_pointcloud(self, update = True):
        """
        Returns the pointcloud dictionary from the class attribute 
        Args:
            update: whether or not to recalculate the pointcloud
        Returns:
            Dict: dictionary of pointcloud points and state vector at time of capture
        """
        #first, calculate the pointcloud
        if update:
            self.calc_ptcloud()
        return self._ptcloudData
    
"""
************************************
Manager objects are contained below
************************************
"""
    
class DepthCamManager:
    def __init__(self, observerManager, obstacleManager, mean, sd):
        """
        Managerial class to manage the depth cameras for N quadrotors
        Args:
            observerManager (ObserverManager): ObserverManager object instance
            obstacleManager (ObstacleManager): manager for the set of NumObs obstacles
            mean (float, optional): Mean for gaussian noise. Defaults to None.
            sd (float, optional): standard deviation for gaussian noise. Defaults to None.
        """
        #store the input parameters
        self.observerManager = observerManager
        self.obstacleManager = obstacleManager
        self.mean = mean
        self.sd = sd

        #create an observer dictionary storing N observer instances
        self.depthCamDict = {}

        #create N lidar objects - one for each turtlebot
        for i in range(self.observerManager.dynamics.N):
            #create an observer with index i
            self.depthCamDict[i] = PlanarDepthCam(i, self.observerManager, self.obstacleManager, mean, sd)

    def get_depth_cam_i(self, i):
        """
        Function to retrieve the ith depth camera object for the quadrotor
        Inputs:
            i (integer): index of the quadrotor whose observer we'd like to retrieve
        """
        return self.depthCamDict[i]
    
class LidarManager:
    def __init__(self, observerManager, mean, sd):
        """
        Managerial class to manage the observers for a system of N turtlebots
        Args:
            observerManager (ObserverManager): ObserverManager object instance
            mean (float, optional): Mean for gaussian noise. Defaults to None.
            sd (float, optional): standard deviation for gaussian noise. Defaults to None.
        """
        #store the input parameters
        self.observerManager = observerManager
        self.mean = mean
        self.sd = sd

        #create an observer dictionary storing N observer instances
        self.lidarDict = {}

        #create N lidar objects - one for each turtlebot
        for i in range(self.observerManager.dynamics.N):
            #create an observer with index i
            self.lidarDict[i] = EgoLidar(i, self.observerManager, mean, sd)

    def get_lidar_i(self, i):
        """
        Function to retrieve the ith observer object for the turtlebot
        Inputs:
            i (integer): index of the turtlebot whose observer we'd like to retrieve
        """
        return self.lidarDict[i]